from __future__ import annotations

import os
import struct
import zlib
import gzip
from typing import Tuple, Any, Dict, Union, Generator, Optional, List, TYPE_CHECKING
import numpy
import time
import re
import threading

import amulet_nbt as nbt

from amulet.api.wrapper.world_format_wrapper import WorldFormatWrapper
from amulet.utils import world_utils
from amulet.utils.format_utils import check_all_exist, load_leveldat
from amulet.api.errors import ChunkDoesNotExist, LevelDoesNotExist, ChunkLoadError
from amulet.api.data_types import ChunkCoordinates, RegionCoordinates

if TYPE_CHECKING:
    from amulet.api.data_types import Dimension

InternalDimension = str


class AnvilRegion:
    region_regex = re.compile(r"r\.(?P<rx>-?\d+)\.(?P<rz>-?\d+)\.mca")

    @staticmethod
    def get_coords(file_path: str) -> Tuple[Union[int, None], Union[int, None]]:
        file_path = os.path.basename(file_path)
        match = AnvilRegion.region_regex.fullmatch(file_path)
        if match is None:
            return None, None
        return int(match.group("rx")), int(match.group("rz"))

    def __init__(self, file_path: str, create=False, mcc=False):
        """
        A class wrapper for a region file
        :param file_path: The file path of the region file
        :param create: bool - if true will create the region from scratch. If false will try loading from disk
        """
        self._file_path = file_path
        self.rx, self.rz = self.get_coords(file_path)
        self._mcc = mcc  # create mcc file if the chunk is greater than 1MiB

        # [dirty, mod_time, data_length, data]  feel free to extend if you want to implement modifying in place and defragging
        # Value is Tuple[save time, compressed bytes]. # The latter can be None if the region file has been unloaded to track which chunks exist.
        # If a chunk does not exist on disk it simply won't exist in this dictionary
        self._chunks: Dict[ChunkCoordinates, Tuple[int, Optional[bytes]]] = {}

        # committed chunks stores chunks that have been committed but not saved to disk. Value is Tuple[save time, compressed bytes].
        # The latter can be None in which case the chunk has been deleted
        self._committed_chunks: Dict[ChunkCoordinates, Tuple[int, Optional[bytes]]] = {}

        self._lock = threading.Lock()
        if create:
            # create the region from scratch.
            self._loaded = True
        else:
            # mark the region to be loaded when needed
            self._loaded = False
            # shallow load the data
            file_size = os.path.getsize(self._file_path)
            if file_size > 4096 * 2:
                with open(self._file_path, "rb") as fp:
                    offsets = numpy.fromfile(fp, dtype=">u4", count=1024).reshape(
                        32, 32
                    )
                    for x in range(32):
                        for z in range(32):
                            offset = offsets[z, x]
                            if offset != 0:
                                self._chunks[(x, z)] = (0, None)

    def all_chunk_coords(self) -> Generator[ChunkCoordinates, None, None]:
        for (cx, cz), (_, chunk_) in self._committed_chunks.items():
            if chunk_ is not None:
                yield cx + self.rx * 32, cz + self.rz * 32
        for (cx, cz), (_, _) in self._chunks.items():
            if (cx, cz) not in self._committed_chunks:
                yield cx + self.rx * 32, cz + self.rz * 32

    def _load(self):
        with self._lock:
            if not self._loaded:
                mode = "rb+" if os.path.isfile(self._file_path) else "wb"
                with open(self._file_path, mode) as fp:
                    # check that the file is a multiple of 4096 bytes and extend if not
                    # TODO: perhaps rewrite this in a way that is more readable
                    file_size = os.path.getsize(self._file_path)
                    if file_size & 0xFFF:
                        file_size = (file_size | 0xFFF) + 1
                        fp.truncate(file_size)

                    # if the length of the region file is 0 extend it to 8KiB TODO (perhaps have some error)
                    if file_size < world_utils.SECTOR_BYTES * 2:
                        file_size = world_utils.SECTOR_BYTES * 2
                        fp.truncate(file_size)

                    # read the file and populate the internal database
                    # self._file_size = file_size

                    fp.seek(0)

                    # offsets = fp.read(world_utils.SECTOR_BYTES)
                    # mod_times = fp.read(world_utils.SECTOR_BYTES)

                    # self._free_sectors = free_sectors = numpy.full(
                    #     file_size // world_utils.SECTOR_BYTES, True, bool
                    # )
                    # self._free_sectors[0:2] = False, False

                    # the first array is made of 3 byte sector offset and 1 byte sector count
                    sectors = (
                        numpy.fromfile(fp, dtype=">u4", count=1024).reshape(32, 32) >> 8
                    )
                    mod_times = numpy.fromfile(fp, dtype=">u4", count=1024).reshape(
                        32, 32
                    )

                    # for offset in offsets:
                    #     sector = offset >> 8
                    #     count = offset & 0xFF
                    #
                    #     for i in range(sector, sector + count):
                    #         if i >= len(free_sectors):
                    #             return False
                    #
                    #         free_sectors[i] = False

                    for cx in range(32):
                        for cz in range(32):
                            sector = sectors[cz, cx]
                            if sector:
                                fp.seek(world_utils.SECTOR_BYTES * sector)
                                # read int value and then read that amount of data
                                buffer_size_bytes: bytes = fp.read(4)
                                buffer_size = struct.unpack(">I", buffer_size_bytes)[0]
                                buffer: bytes = fp.read(buffer_size)

                                if buffer:
                                    if buffer[0] & 128:  # if the "external" bit is set
                                        if self._mcc:
                                            mcc_path = os.path.join(
                                                os.path.dirname(self._file_path),
                                                f"c.{cx + self.rx * 32}.{cz + self.rz * 32}.mcc",
                                            )
                                            if os.path.isfile(mcc_path):
                                                with open(mcc_path, "rb") as f:
                                                    buffer = (
                                                        bytes([buffer[0] & 127])
                                                        + f.read()
                                                    )
                                            else:
                                                # the external flag was set but the external file cannot be found. Continue as if the chunk does not exist.
                                                continue
                                        else:
                                            # External bit set but this version cannot handle mcc files. Continue as if the chunk does not exist.
                                            continue

                                    self._chunks[(cx, cz)] = (
                                        mod_times[cz, cx],
                                        buffer,
                                    )

                self._loaded = True

    def unload(self):
        for key in self._chunks.keys():
            self._chunks[key] = (0, None)
        self._loaded = False

    def save(self):
        if self._committed_chunks:
            self._load()
            if self._mcc:
                mcc_chunks = {(cx, cz) for cx in range(32) for cz in range(32)}
            else:
                mcc_chunks = set()
            for key, val in self._committed_chunks.items():
                if val[1] is not None:
                    if self._mcc or len(val[1]) <= 2 ** 20 - 4:
                        self._chunks[key] = val
                elif key in self._chunks:
                    del self._chunks[key]
            self._committed_chunks.clear()
            offsets = numpy.zeros(1024, dtype=">u4")
            mod_times = numpy.zeros(1024, dtype=">u4")
            offset = 2
            data = []
            for (cx, cz), (mod_time, buffer) in self._chunks.items():
                if buffer is not None:
                    index = cx + (cz << 5)
                    buffer_size = len(buffer)
                    if (
                        buffer_size > 2 ** 20 - 4
                    ):  # if mcc is false the chunks that are too large should have already been removed.
                        mcc_chunks.remove((cx, cz))
                        with open(
                            os.path.join(
                                os.path.dirname(self._file_path),
                                f"c.{cx + self.rx * 32}.{cz + self.rz * 32}.mcc",
                            ),
                            "wb",
                        ) as f:
                            f.write(buffer[1:])
                        buffer = bytes([buffer[0] | 128])  # set the external flag
                        buffer_size = 1

                    sector_count = ((buffer_size + 4 | 0xFFF) + 1) >> 12
                    offsets[index] = (offset << 8) + sector_count
                    mod_times[index] = mod_time
                    data.append(
                        struct.pack(">I", buffer_size)
                        + buffer
                        + b"\x00" * ((sector_count << 12) - buffer_size - 4)
                    )
                    offset += sector_count
            os.makedirs(os.path.dirname(self._file_path), exist_ok=True)
            with open(self._file_path, "wb") as fp:
                fp.write(
                    struct.pack(">1024I", *offsets)
                )  # there is probably a prettier way of doing this
                fp.write(
                    struct.pack(">1024I", *mod_times)
                )  # but I could not work it out with Numpy
                fp.write(b"".join(data))

            # remove orphaned mcc files
            for cx, cz in mcc_chunks:
                mcc_path = os.path.join(
                    os.path.dirname(self._file_path),
                    f"c.{cx + self.rx * 32}.{cz + self.rz * 32}.mcc",
                )
                if os.path.isfile(mcc_path):
                    os.remove(mcc_path)

    def get_chunk_data(self, cx: int, cz: int) -> nbt.NBTFile:
        """Get chunk data. Coords are in region space."""
        if (cx, cz) in self._committed_chunks:
            # if the chunk exists in the committed but unsaved chunks return that
            data = self._committed_chunks[(cx, cz)][1]
            if data is not None:
                compress_type, data = data[0], data[1:]
                return self._decompress(compress_type, data)
        elif (cx, cz) in self._chunks:
            # otherwise if the chunk exists in the main database return that
            self._load()
            data: bytes = self._chunks[(cx, cz)][1]
            compress_type, data = data[0], data[1:]
            if data:
                return self._decompress(compress_type, data)

        raise ChunkDoesNotExist

    def put_chunk_data(self, cx: int, cz: int, data: nbt.NBTFile):
        """compress the data and put it in the class database"""
        bytes_data = self._compress(data)
        self._committed_chunks[(cx, cz)] = (int(time.time()), bytes_data)

    def delete_chunk_data(self, cx: int, cz: int):
        self._committed_chunks[(cx, cz)] = (0, None)

    @staticmethod
    def _compress(data: nbt.NBTFile) -> bytes:
        """Convert an NBTFile into a compressed bytes object"""
        data = data.save_to(compressed=False)
        return b"\x02" + zlib.compress(data)

    @staticmethod
    def _decompress(compress_type: int, data: bytes) -> nbt.NBTFile:
        """Convert a bytes object into an NBTFile"""
        if compress_type == world_utils.VERSION_GZIP:
            return nbt.load(buffer=gzip.decompress(data), compressed=False)
        elif compress_type == world_utils.VERSION_DEFLATE:
            return nbt.load(buffer=zlib.decompress(data), compressed=False)
        raise ChunkLoadError(f"Invalid compression type {compress_type}")


class AnvilLevelManager:
    level_regex = re.compile(r"DIM(?P<level>-?\d+)")

    def __init__(self, directory: str, mcc=False):
        self._directory = directory
        self._regions: Dict[RegionCoordinates, AnvilRegion] = {}
        self._mcc = mcc

        # shallow load all of the existing region classes
        region_dir = os.path.join(self._directory, "region")
        if os.path.isdir(region_dir):
            for region_file_name in os.listdir(region_dir):
                rx, rz = AnvilRegion.get_coords(region_file_name)
                if rx is None:
                    continue
                self._regions[(rx, rz)] = AnvilRegion(
                    os.path.join(self._directory, "region", region_file_name),
                    mcc=self._mcc,
                )

    def all_chunk_coords(self) -> Generator[ChunkCoordinates, None, None]:
        for region in self._regions.values():
            yield from region.all_chunk_coords()

    def save(self, unload=True):
        # use put_chunk_data to actually upload modified chunks
        # run this to push those changes to disk

        for region in self._regions.values():
            region.save()
            if unload:
                region.unload()

    def close(self):
        pass

    def unload(self):
        for region in self._regions.values():
            region.unload()

    def get_chunk_data(self, cx: int, cz: int) -> nbt.NBTFile:
        """Get an NBTFile of a chunk from the database.
        Will raise ChunkDoesNotExist if the region or chunk does not exist
        """
        # get the region key
        return self._get_region(cx, cz).get_chunk_data(cx & 0x1F, cz & 0x1F)

    def _get_region(self, cx: int, cz: int, create=False) -> AnvilRegion:
        key = rx, rz = world_utils.chunk_coords_to_region_coords(cx, cz)
        if key in self._regions:
            return self._regions[key]

        if create:
            file_path = os.path.join(self._directory, "region", f"r.{rx}.{rz}.mca")
            self._regions[key] = AnvilRegion(file_path, True, mcc=self._mcc)
        else:
            raise ChunkDoesNotExist

        return self._regions[key]

    def put_chunk_data(self, cx: int, cz: int, data: nbt.NBTFile):
        """pass data to the region file class"""
        # get the region key
        self._get_region(cx, cz, create=True).put_chunk_data(cx & 0x1F, cz & 0x1F, data)

    def delete_chunk(self, cx: int, cz: int):
        try:
            self._get_region(cx, cz).delete_chunk_data(cx & 0x1F, cz & 0x1F)
        except ChunkDoesNotExist:
            pass


class AnvilFormat(WorldFormatWrapper):
    def __init__(self, directory: str):
        super().__init__(directory)
        self.root_tag: nbt.NBTFile = nbt.NBTFile()
        self._load_level_dat()
        self._levels: Dict[InternalDimension, AnvilLevelManager] = {}
        self._dimension_name_map: Dict["Dimension", InternalDimension] = {}
        self._mcc_support: Optional[bool] = None
        self._lock = None

    def _load_level_dat(self):
        """Load the level.dat file and check the image file"""
        self.root_tag = nbt.load(os.path.join(self.path, "level.dat"))
        if os.path.isfile(os.path.join(self.path, "icon.png")):
            self._world_image_path = os.path.join(self.path, "icon.png")
        else:
            self._world_image_path = self._missing_world_icon

    @staticmethod
    def is_valid(directory) -> bool:
        """
        Returns whether this format is able to load a given world.

        :param directory: The path to the root of the world to load.
        :return: True if the world can be loaded by this format, False otherwise.
        """
        if not check_all_exist(directory, "level.dat"):
            return False

        try:
            level_dat_root = load_leveldat(directory)
        except:
            return False

        if "Data" not in level_dat_root:
            return False

        if "FML" in level_dat_root:
            return False

        return True

    @property
    def platform(self) -> str:
        """Platform string"""
        return "java"

    def _get_version(self) -> int:
        return (
            self.root_tag.get("Data", nbt.TAG_Compound())
            .get("DataVersion", nbt.TAG_Int(-1))
            .value
        )

    @property
    def world_name(self) -> str:
        """The name of the world"""
        return self.root_tag["Data"]["LevelName"].value

    @world_name.setter
    def world_name(self, value: str):
        self.root_tag["Data"]["LevelName"] = nbt.TAG_String(value)

    @property
    def last_played(self) -> int:
        return self.root_tag["Data"]["LastPlayed"].value

    @property
    def game_version_string(self) -> str:
        try:
            return f'Java {self.root_tag["Data"]["Version"]["Name"].value}'
        except Exception:
            return f"Java Unknown Version"

    @property
    def dimensions(self) -> List["Dimension"]:
        """A list of all the levels contained in the world"""
        return list(self._dimension_name_map.keys())

    def register_dimension(
        self,
        dimension_internal: InternalDimension,
        dimension_name: Optional["Dimension"] = None,
    ):
        """
        Register a new dimension.
        :param dimension_internal: The internal representation of the dimension
        :param dimension_name: The name of the dimension shown to the user
        :return:
        """
        if dimension_name is None:
            dimension_name: "Dimension" = dimension_internal

        if dimension_internal:
            path = os.path.join(self.path, dimension_internal)
        else:
            path = self.path

        if (
            dimension_internal not in self._levels
            and dimension_name not in self._dimension_name_map
        ):
            self._levels[dimension_internal] = AnvilLevelManager(
                path, mcc=self._mcc_support
            )
            self._dimension_name_map[dimension_name] = dimension_internal

    def _get_interface_key(self, raw_chunk_data) -> Tuple[str, int]:
        return self.platform, raw_chunk_data.get("DataVersion", nbt.TAG_Int(-1)).value

    def _reload_world(self):
        # reload the level.dat in case it has changed
        self._load_level_dat()

        # create the session.lock file (this has mostly been lifted from MCEdit)
        self._lock = int(time.time() * 1000)
        with open(os.path.join(self.path, "session.lock"), "wb") as f:
            f.write(struct.pack(">Q", self._lock))
            f.flush()
            os.fsync(f.fileno())

        self._mcc_support = (
            self.version > 2203
        )  # the real number might actually be lower

        # load all the levels
        self.register_dimension("", "overworld")
        self.register_dimension("DIM-1", "nether")
        self.register_dimension("DIM1", "end")

        for dir_name in os.listdir(self.path):
            level_path = os.path.join(self.path, dir_name)
            if os.path.isdir(level_path) and dir_name.startswith("DIM"):
                if AnvilLevelManager.level_regex.fullmatch(dir_name) is None:
                    continue
                self.register_dimension(dir_name)

    def open(self):
        """Open the database for reading and writing"""
        self._reload_world()

    @property
    def has_lock(self) -> bool:
        """Verify that the world database can be read and written"""
        try:
            with open(os.path.join(self.path, "session.lock"), "rb") as f:
                return struct.unpack(">Q", f.read(8))[0] == self._lock
        except Exception:
            return False

    def save(self):
        """Save the data back to the disk database"""
        self._verify_has_lock()
        for level in self._levels.values():
            level.save()
        # TODO: save other world data
        self._changed = False

    def close(self):
        """Close the disk database"""
        pass

    def unload(self):
        """Unload data stored in the Format class"""
        for level in self._levels.values():
            level.unload()

    def _has_dimension(self, dimension: "Dimension"):
        return (
            dimension in self._dimension_name_map
            and self._dimension_name_map[dimension] in self._levels
        )

    def _get_dimension(self, dimension: "Dimension"):
        self._verify_has_lock()
        if self._has_dimension(dimension):
            return self._levels[self._dimension_name_map[dimension]]
        else:
            raise LevelDoesNotExist(dimension)

    def all_chunk_coords(
        self, dimension: "Dimension"
    ) -> Generator[ChunkCoordinates, None, None]:
        """A generator of all chunk coords in the given dimension"""
        if self._has_dimension(dimension):
            yield from self._get_dimension(dimension).all_chunk_coords()

    def delete_chunk(self, cx: int, cz: int, dimension: "Dimension"):
        """Delete a chunk from a given dimension"""
        if self._has_dimension(dimension):
            self._get_dimension(dimension).delete_chunk(cx, cz)

    def _put_raw_chunk_data(
        self, cx: int, cz: int, data: Any, dimension: "Dimension", *args
    ):
        """
        Actually stores the data from the interface to disk.
        """
        self._get_dimension(dimension).put_chunk_data(cx, cz, data)

    def _get_raw_chunk_data(
        self, cx: int, cz: int, dimension: "Dimension", *args
    ) -> nbt.NBTFile:
        """
        Return the interface key and data to interface with given chunk coordinates.

        :param cx: The x coordinate of the chunk.
        :param cz: The z coordinate of the chunk.
        :return: The interface key for the get_interface method and the data to interface with.
        """
        return self._get_dimension(dimension).get_chunk_data(cx, cz)


FORMAT_CLASS = AnvilFormat

if __name__ == "__main__":
    import sys

    world_path = sys.argv[1]
    world = AnvilLevelManager(world_path)
    chunk = world.get_chunk_data(0, 0)
    print(chunk)
    world.put_chunk_data(0, 0, chunk)
    world.save()



from __future__ import annotations

from amulet.world_interface.chunk.interfaces.leveldb.leveldb_19.interface import (
    LevelDB19Interface,
)


class LevelDB20Interface(LevelDB19Interface):
    def __init__(self):
        super().__init__()

        self.features["chunk_version"] = 20


INTERFACE_CLASS = LevelDB20Interface



from __future__ import annotations

from amulet.world_interface.chunk.interfaces.leveldb.leveldb_8.interface import (
    LevelDB8Interface,
)


class LevelDB9Interface(LevelDB8Interface):
    def __init__(self):
        super().__init__()

        self.features["chunk_version"] = 9
        self.features[
            "entity_format"
        ] = "namespace-str-identifier"  # "int-id" is present until at least v7. Not sure which was present for v8


INTERFACE_CLASS = LevelDB9Interface



from __future__ import annotations

from typing import TYPE_CHECKING
import numpy

from amulet.api.structure import Structure
from amulet.api.errors import ChunkLoadError, ChunkDoesNotExist
from amulet.api.chunk import Chunk
from amulet.api.data_types import Dimension, BlockCoordinates, FloatTriplet
from amulet.utils.matrix import transform_matrix

if TYPE_CHECKING:
    from amulet.api.world import World


def paste(
    world: "World",
    dimension: Dimension,
    structure: Structure,
    location: BlockCoordinates,
    scale: FloatTriplet,
    rotation: FloatTriplet,
    copy_air=True,
    copy_water=True,
    copy_lava=True,
):
    for _ in paste_iter(
        world,
        dimension,
        structure,
        location,
        scale,
        rotation,
        copy_air,
        copy_water,
        copy_lava,
    ):
        pass


def paste_iter(
    world: "World",
    dimension: Dimension,
    structure: Structure,
    location: BlockCoordinates,
    scale: FloatTriplet,
    rotation: FloatTriplet,
    copy_air=True,
    copy_water=True,
    copy_lava=True,
):
    gab = numpy.vectorize(world.palette.get_add_block, otypes=[numpy.uint32])
    lut = gab(structure.palette.blocks())
    filtered_mode = not all([copy_air, copy_lava, copy_water])
    filtered_blocks = []
    if not copy_air:
        filtered_blocks.append("universal_minecraft:air")
    if not copy_water:
        filtered_blocks.append("universal_minecraft:water")
    if not copy_lava:
        filtered_blocks.append("universal_minecraft:lava")
    if filtered_mode:
        paste_blocks = numpy.array(
            [
                any(
                    sub_block.namespaced_name not in filtered_blocks
                    for sub_block in block.block_tuple
                )
                for block in structure.palette.blocks()
            ]
        )
    else:
        paste_blocks = None

    rotation_point = numpy.floor(
        (structure.selection.max + structure.selection.min) / 2
    ).astype(int)

    if any(rotation) or any(s != 1 for s in scale):
        yield 0, "Rotating!"
        transformed_structure = yield from structure.transform_iter(scale, rotation)
        rotation_point = (
            numpy.matmul(
                transform_matrix(
                    (0, 0, 0), scale, -numpy.radians(numpy.flip(rotation)), "zyx"
                ),
                numpy.array([*rotation_point, 1]),
            )
            .T[:3]
            .round()
            .astype(int)
        )
    else:
        transformed_structure = structure

    offset = location - rotation_point
    moved_min_location = transformed_structure.selection.min + offset

    iter_count = len(
        list(
            transformed_structure.get_moved_chunk_slices(
                moved_min_location, generate_non_exists=True
            )
        )
    )
    count = 0

    yield 0, "Pasting!"
    for (
        src_chunk,
        src_slices,
        src_box,
        (dst_cx, dst_cz),
        dst_slices,
        dst_box,
    ) in transformed_structure.get_moved_chunk_slices(
        moved_min_location, generate_non_exists=True
    ):
        print(src_chunk)
        try:
            dst_chunk = world.get_chunk(dst_cx, dst_cz, dimension)
        except ChunkDoesNotExist:
            dst_chunk = Chunk(dst_cx, dst_cz)
            world.put_chunk(dst_chunk, dimension)
        except ChunkLoadError:
            continue
        remove_block_entities = []
        for block_entity_location in dst_chunk.block_entities.keys():
            if block_entity_location in dst_box:
                if copy_air:
                    remove_block_entities.append(block_entity_location)
                else:
                    chunk_block_entity_location = (
                        numpy.array(block_entity_location) - offset
                    )
                    chunk_block_entity_location[[0, 2]] %= 16
                    if paste_blocks[
                        src_chunk.blocks[tuple(chunk_block_entity_location)]
                    ]:
                        remove_block_entities.append(block_entity_location)
        for block_entity_location in remove_block_entities:
            del dst_chunk.block_entities[block_entity_location]
        for block_entity_location, block_entity in src_chunk.block_entities.items():
            if block_entity_location in src_box:
                dst_chunk.block_entities.insert(
                    block_entity.new_at_location(*offset + block_entity_location)
                )

        if not copy_air:
            mask = paste_blocks[src_chunk.blocks[src_slices]]
            dst_chunk.blocks[dst_slices][mask] = lut[src_chunk.blocks[src_slices]][mask]
        else:
            dst_chunk.blocks[dst_slices] = lut[src_chunk.blocks[src_slices]]
        dst_chunk.changed = True

        count += 1
        yield count / iter_count



import copy
from typing import Dict, Union, Generator, Tuple, Optional, List
import itertools
import numpy

from .selection import SelectionGroup, SelectionBox
from .base_structure import BaseStructure
from .world import World
from .chunk import Chunk
from .block import Block
from amulet.api.registry import BlockManager
from .errors import ChunkDoesNotExist
from amulet.api.data_types import ChunkCoordinates, Dimension, FloatTriplet
from ..utils.world_utils import block_coords_to_chunk_coords
from amulet.utils.matrix import transform_matrix


class StructureCache:
    """A class for storing and accessing structure objects"""

    def __init__(self):
        self._structure_buffer: List[Structure] = []

    def add_structure(self, structure: "Structure"):
        """Add a structure to the cache"""
        assert isinstance(
            structure, Structure
        ), "structure given is not a Structure instance"
        self._structure_buffer.append(structure)

    def get_structure(self, index=-1) -> "Structure":
        """Get a structure from the cache. Default last."""
        return self._structure_buffer[index]

    def pop_structure(self, index=-1) -> "Structure":
        """Get and remove a structure from the cache. Default last."""
        return self._structure_buffer.pop(index)

    def __len__(self) -> int:
        """Get the number of structures in the cache"""
        return len(self._structure_buffer)

    def clear(self):
        """Empty the cache."""
        self._structure_buffer.clear()


structure_cache = StructureCache()


class Structure(BaseStructure):
    def __init__(
        self,
        chunks: Dict[ChunkCoordinates, Chunk],
        palette: BlockManager,
        selection: SelectionGroup,
        chunk_size: Tuple[int, int, int] = (16, 256, 16),
    ):
        self._chunk_cache = chunks
        self._palette = palette
        self._selection = selection
        self._chunk_size = chunk_size

    @property
    def chunk_size(self) -> Tuple[int, int, int]:
        return self._chunk_size

    @property
    def palette(self) -> BlockManager:
        return self._palette

    @property
    def selection(self) -> SelectionGroup:
        return self._selection

    @classmethod
    def from_world(cls, world: World, selection: SelectionGroup, dimension: Dimension):
        data = {}
        block_palette = BlockManager()
        for chunk, _ in world.get_chunk_boxes(selection, dimension):
            if chunk.coordinates not in data:
                data[chunk.coordinates] = chunk = copy.deepcopy(chunk)
                chunk.block_palette = block_palette
        return cls(data, block_palette, copy.deepcopy(selection), world.chunk_size)

    def get_chunk(self, cx: int, cz: int) -> Chunk:
        if (cx, cz) in self._chunk_cache:
            return self._chunk_cache[(cx, cz)]
        else:
            raise ChunkDoesNotExist

    def get_block(self, x: int, y: int, z: int) -> Block:
        cx, cz = block_coords_to_chunk_coords(x, z, chunk_size=self.sub_chunk_size)
        offset_x, offset_z = x - 16 * cx, z - 16 * cz

        chunk = self.get_chunk(cx, cz)
        block = chunk.blocks[offset_x, y, offset_z]
        return self._palette[block]

    def get_chunk_boxes(
        self,
        selection: Optional[Union[SelectionGroup, SelectionBox]] = None,
        generate_non_exists: bool = False,
    ) -> Generator[Tuple[Chunk, SelectionBox], None, None]:
        """Given a selection will yield chunks and SubSelectionBoxes into that chunk

        :param selection: SelectionGroup or SelectionBox into the world
        :param generate_non_exists: Generate empty read-only chunks if the chunk does not exist.
        """
        if selection is None:
            selection = self._selection
        else:
            if isinstance(selection, SelectionBox):
                selection = SelectionGroup([selection])
            # TODO: handle the fact the the selection is not at the origin
            selection = self.selection.intersection(selection)
        selection: SelectionGroup
        for box in selection.selection_boxes:
            first_chunk = block_coords_to_chunk_coords(
                box.min_x, box.min_z, chunk_size=self.sub_chunk_size
            )
            last_chunk = block_coords_to_chunk_coords(
                box.max_x - 1, box.max_z - 1, chunk_size=self.sub_chunk_size
            )
            for cx, cz in itertools.product(
                range(first_chunk[0], last_chunk[0] + 1),
                range(first_chunk[1], last_chunk[1] + 1),
            ):
                try:
                    chunk = self.get_chunk(cx, cz)
                except ChunkDoesNotExist:
                    if generate_non_exists:
                        chunk = Chunk(cx, cz)
                        chunk.block_palette = self.palette
                    else:
                        continue

                yield chunk, box.intersection(self._chunk_box(cx, cz))

    def get_chunk_slices(
        self, selection: Optional[Union[SelectionGroup, SelectionBox]] = None
    ) -> Generator[Tuple[Chunk, Tuple[slice, slice, slice], SelectionBox], None, None]:
        """Given a selection will yield chunks and slices into that chunk

        :param selection: SelectionGroup or SelectionBox into the world
        Usage:
        for chunk, slice in world.get_chunk_slices(selection):
            chunk.blocks[slice] = ...
        """
        for chunk, box in self.get_chunk_boxes(selection):
            slices = box.chunk_slice(chunk.cx, chunk.cz, self.sub_chunk_size)
            yield chunk, slices, box

    def get_moved_chunk_slices(
        self,
        destination_origin: Tuple[int, int, int],
        selection: Optional[Union[SelectionGroup, SelectionBox]] = None,
        destination_chunk_shape: Optional[Tuple[int, int, int]] = None,
        generate_non_exists: bool = False,
    ) -> Generator[
        Tuple[
            Chunk,
            Tuple[slice, slice, slice],
            SelectionBox,
            ChunkCoordinates,
            Tuple[slice, slice, slice],
            SelectionBox,
        ],
        None,
        None,
    ]:
        """Iterate over a selection and return slices into the source object and destination object
        given the origin of the destination. When copying a selection to a new area the slices will
        only be equal if the offset is a multiple of the chunk size. This will rarely be the case
        so the slices need to be split up into parts that intersect a chunk in the source and destination.
        :param destination_origin: The location where the minimum point of self.selection will end up
        :param selection: An optional selection. The overlap of this and self.selection will be used
        :param destination_chunk_shape: the chunk shape of the destination object (defaults to self.chunk_size)
        :param generate_non_exists: Generate empty read-only chunks if the chunk does not exist.
        :return:
        """
        if destination_chunk_shape is None:
            destination_chunk_shape = self.chunk_size

        if selection is None:
            selection = self.selection
        else:
            # TODO: handle the fact the the selection is not at the origin
            selection = self.selection.intersection(selection)
        # the offset from self.selection to the destination location
        offset = numpy.subtract(destination_origin, self.selection.min, dtype=numpy.int)
        for chunk, box in self.get_chunk_boxes(selection, generate_non_exists):
            dst_full_box = SelectionBox(
                offset + box.min,
                offset + box.max,
            )

            first_chunk = block_coords_to_chunk_coords(
                dst_full_box.min_x,
                dst_full_box.min_z,
                chunk_size=destination_chunk_shape[0],
            )
            last_chunk = block_coords_to_chunk_coords(
                dst_full_box.max_x - 1,
                dst_full_box.max_z - 1,
                chunk_size=destination_chunk_shape[0],
            )
            for cx, cz in itertools.product(
                range(first_chunk[0], last_chunk[0] + 1),
                range(first_chunk[1], last_chunk[1] + 1),
            ):
                chunk_box = self._chunk_box(cx, cz, destination_chunk_shape)
                dst_box = chunk_box.intersection(dst_full_box)
                src_box = SelectionBox(-offset + dst_box.min, -offset + dst_box.max)
                src_slices = src_box.chunk_slice(
                    chunk.cx, chunk.cz, self.sub_chunk_size
                )
                dst_slices = dst_box.chunk_slice(cx, cz, self.sub_chunk_size)
                yield chunk, src_slices, src_box, (cx, cz), dst_slices, dst_box

    def transform(self, scale: FloatTriplet, rotation: FloatTriplet) -> "Structure":
        """
        creates a new transformed Structure class.
        :param scale: scale factor multiplier in the x, y and z directions
        :param rotation: rotation in degrees for pitch (y), yaw (z) and roll (x)
        :return:
        """
        iter_ = self.transform_iter(scale, rotation)
        try:
            while True:
                next(iter_)
        except StopIteration as e:
            return e.value

    def transform_iter(
        self, scale: FloatTriplet, rotation: FloatTriplet
    ) -> Generator[int, None, "Structure"]:
        """
        creates a new transformed Structure class.
        :param scale: scale factor multiplier in the x, y and z directions
        :param rotation: rotation in degrees for pitch (y), yaw (z) and roll (x)
        :return:
        """
        block_palette = copy.deepcopy(self.palette)
        rotation_radians = -numpy.flip(numpy.radians(rotation))
        selection = self.selection.transform(scale, rotation_radians)
        transform = transform_matrix((0, 0, 0), scale, rotation_radians, "zyx")
        inverse_transform = numpy.linalg.inv(transform)

        chunks: Dict[ChunkCoordinates, Chunk] = {}

        volume = sum([box.volume for box in selection.selection_boxes])
        index = 0

        # TODO: find a way to do this without doing it block by block
        for box in selection.selection_boxes:
            coords = list(box.blocks())
            coords_array = numpy.ones((len(coords), 4), dtype=numpy.float)
            coords_array[:, :3] = coords
            coords_array[:, :3] += 0.5
            original_coords = (
                numpy.floor(numpy.matmul(inverse_transform, coords_array.T))
                .astype(int)
                .T[:, :3]
            )
            for (x, y, z), (ox, oy, oz) in zip(coords, original_coords):
                cx, cz = chunk_key = (x >> 4, z >> 4)
                if chunk_key in chunks:
                    chunk = chunks[chunk_key]
                else:
                    chunk = chunks[chunk_key] = Chunk(cx, cz)
                    chunk.block_palette = block_palette
                try:
                    chunk.blocks[x % 16, y, z % 16] = self.get_chunk(
                        ox >> 4, oz >> 4
                    ).blocks[ox % 16, oy, oz % 16]
                except ChunkDoesNotExist:
                    pass
                yield index / volume
                index += 1

        return Structure(chunks, block_palette, selection, self.chunk_size)



from __future__ import annotations

import os
from amulet.world_interface.loader import Loader

SUPPORTED_FORMAT_VERSION = 0
SUPPORTED_META_VERSION = 0

FORMATS_DIRECTORY = os.path.dirname(__file__)

loader = Loader(
    "format",
    FORMATS_DIRECTORY,
    SUPPORTED_META_VERSION,
    SUPPORTED_FORMAT_VERSION,
    create_instance=False,
)

if __name__ == "__main__":
    import time

    print(loader.get_all())
    time.sleep(1)
    loader.reload()
    print(loader.get_all())



from __future__ import annotations

from amulet.world_interface.chunk.interfaces.leveldb.leveldb_0.interface import (
    LevelDB0Interface,
)


class LevelDB1Interface(LevelDB0Interface):
    def __init__(self):
        super().__init__()

        self.features["chunk_version"] = 1


INTERFACE_CLASS = LevelDB1Interface



from __future__ import annotations

from amulet.world_interface.chunk.interfaces.leveldb.leveldb_11.interface import (
    LevelDB11Interface,
)


class LevelDB12Interface(LevelDB11Interface):
    def __init__(self):
        super().__init__()

        self.features["chunk_version"] = 12


INTERFACE_CLASS = LevelDB12Interface



from __future__ import annotations

from amulet.world_interface.chunk.interfaces.leveldb.leveldb_15.interface import (
    LevelDB15Interface,
)


class LevelDB16Interface(LevelDB15Interface):
    def __init__(self):
        super().__init__()

        self.features["chunk_version"] = 16


INTERFACE_CLASS = LevelDB16Interface



from __future__ import annotations

import os
from typing import Union, Tuple, IO, List, Generator, Dict
import numpy

import amulet_nbt

from amulet.api.selection import SelectionBox
from amulet.api.data_types import ChunkCoordinates, BlockCoordinates, AnyNDArray
from amulet.api.errors import ChunkDoesNotExist
from amulet.utils.numpy_helpers import brute_sort_objects_no_hash


class MCStructureChunk:
    __slots__ = (
        "selection",
        "blocks",
        "palette",
        "shape",
        "entities",
        "block_entities",
    )

    def __init__(
        self,
        selection: SelectionBox,
        blocks: numpy.ndarray,
        palette: AnyNDArray,
        block_entities: List[amulet_nbt.TAG_Compound],
        entities: List[amulet_nbt.TAG_Compound],
    ):
        assert isinstance(blocks, numpy.ndarray)
        assert (
            selection.shape == blocks.shape
        ), "blocks shape does not match the specified section shape"
        self.selection = selection
        self.blocks = blocks
        self.palette = palette
        self.block_entities = block_entities
        self.entities = entities

    def __eq__(self, other):
        return (
            isinstance(other, MCStructureChunk)
            and self.selection == other.selection
            and self.shape == other.shape
            and numpy.array_equal(self.blocks, other.blocks)
            and self.entities == other.entities
            and self.block_entities == other.block_entities
        )

    @property
    def location(self) -> BlockCoordinates:
        return self.selection.min


class MCStructureReader:
    def __init__(self, path_or_buffer: Union[str, IO]):
        if isinstance(path_or_buffer, str):
            assert path_or_buffer.endswith(
                ".mcstructure"
            ), "File selected is not a .mcstructure file"
            assert os.path.isfile(
                path_or_buffer
            ), f"There is no mcstructure file at path {path_or_buffer}"
            mcstructure = amulet_nbt.load(path_or_buffer, little_endian=True)
        else:
            assert hasattr(path_or_buffer, "read"), "Object does not have a read method"
            mcstructure = amulet_nbt.load(buffer=path_or_buffer, little_endian=True)

        self._chunks: Dict[
            ChunkCoordinates,
            Tuple[
                SelectionBox,
                numpy.ndarray,
                AnyNDArray,
                List[amulet_nbt.TAG_Compound],
                List[amulet_nbt.TAG_Compound],
            ],
        ] = {}
        if mcstructure["format_version"].value == 1:
            min_point = numpy.array(
                tuple(c.value for c in mcstructure["structure_world_origin"])
            )
            max_point = min_point + tuple(c.value for c in mcstructure["size"])
            self._selection = SelectionBox(min_point, max_point)
            blocks_array: numpy.ndarray = numpy.array(
                [
                    [b.value for b in layer]
                    for layer in mcstructure["structure"]["block_indices"]
                ],
                dtype=numpy.int32,
            ).reshape(
                (len(mcstructure["structure"]["block_indices"]), *self._selection.shape)
            )

            palette_key = list(mcstructure["structure"]["palette"].keys())[
                0
            ]  # find a way to do this based on user input
            block_palette = list(
                mcstructure["structure"]["palette"][palette_key]["block_palette"]
            )

            for cx, cz in self._selection.chunk_locations():
                chunk_box = SelectionBox.create_chunk_box(cx, cz).intersection(
                    self._selection
                )
                array_slice = (slice(None),) + chunk_box.create_moved_box(
                    self._selection.min, subtract=True
                ).slice
                chunk_blocks_: numpy.ndarray = blocks_array[array_slice]
                chunk_palette_indexes, chunk_blocks = numpy.unique(
                    chunk_blocks_.reshape((chunk_blocks_.shape[0], -1)).T,
                    return_inverse=True,
                    axis=0,
                )
                chunk_blocks = chunk_blocks.reshape(chunk_blocks_.shape[1:])

                chunk_palette = numpy.empty(len(chunk_palette_indexes), dtype=object)
                for palette_index, indexes in enumerate(chunk_palette_indexes):
                    chunk_palette[palette_index] = tuple(
                        block_palette[index] for index in indexes if index >= 0
                    )

                self._chunks[(cx, cz)] = (
                    chunk_box,
                    chunk_blocks,
                    chunk_palette,
                    [],
                    [],
                )

            block_entities = {
                int(key): val["block_entity_data"]
                for key, val in mcstructure["structure"]["palette"][palette_key][
                    "block_position_data"
                ].items()
                if "block_entity_data" in val
            }
            for location, block_entity in block_entities.items():
                if all(key in block_entity for key in "xyz"):
                    x, y, z = (
                        block_entity["x"].value,
                        block_entity["y"].value,
                        block_entity["z"].value,
                    )
                    cx, cz = x >> 4, z >> 4
                    if (cx, cz) in self._chunks and (x, y, z) in self._chunks[(cx, cz)][
                        0
                    ]:
                        self._chunks[(cx, cz)][3].append(block_entity)

            entities = list(mcstructure["structure"]["entities"])
            for entity in entities:
                if "Pos" in entity:
                    x, y, z = (
                        entity["Pos"][0].value,
                        entity["Pos"][1].value,
                        entity["Pos"][2].value,
                    )
                    cx, cz = numpy.floor([x, z]).astype(numpy.int) >> 4
                    if (cx, cz) in self._chunks and (x, y, z) in self._chunks[(cx, cz)][
                        0
                    ]:
                        self._chunks[(cx, cz)][4].append(entity)

        else:
            raise Exception(
                f"mcstructure file with format_version=={mcstructure['format_version'].value} cannot be read"
            )

    def read(self, cx: int, cz: int):
        if (cx, cz) in self._chunks:
            return MCStructureChunk(*self._chunks[(cx, cz)])
        else:
            raise ChunkDoesNotExist

    @property
    def selection(self) -> SelectionBox:
        return self._selection

    @property
    def chunk_coords(self) -> Generator[ChunkCoordinates, None, None]:
        yield from self._chunks.keys()

    def close(self):
        pass


class MCStructureWriter:
    def __init__(self, path_or_buffer: str, selection: SelectionBox):
        self._path_or_buffer = path_or_buffer
        self._selection = selection

        self._data = amulet_nbt.NBTFile(
            amulet_nbt.TAG_Compound(
                {
                    "format_version": amulet_nbt.TAG_Int(1),
                    "structure_world_origin": amulet_nbt.TAG_List(
                        [
                            amulet_nbt.TAG_Int(selection.min_x),
                            amulet_nbt.TAG_Int(selection.min_y),
                            amulet_nbt.TAG_Int(selection.min_z),
                        ]
                    ),
                    "size": amulet_nbt.TAG_List(
                        [
                            amulet_nbt.TAG_Int(selection.size_x),
                            amulet_nbt.TAG_Int(selection.size_y),
                            amulet_nbt.TAG_Int(selection.size_z),
                        ]
                    ),
                }
            )
        )

        self._entities = []
        self._block_entities = []
        self._blocks = numpy.zeros(
            selection.shape, dtype=numpy.uint32
        )  # only 12 bits are actually used at most
        self._palette: List[AnyNDArray] = []
        self._palette_len = 0

    def write(self, section: MCStructureChunk):
        if section.selection.intersects(self._selection):
            box = section.selection.create_moved_box(self._selection.min, subtract=True)
            self._blocks[box.slice] = section.blocks + self._palette_len
            self._palette.append(section.palette)
            self._palette_len += len(section.palette)
            self._block_entities += section.block_entities
            self._entities += section.entities

    def close(self):
        compact_palette, lut = brute_sort_objects_no_hash(
            numpy.concatenate(self._palette)
        )
        self._blocks = lut[self._blocks].ravel()
        block_palette = []
        block_palette_indices = []
        for block_list in compact_palette:
            indexed_block = [-1] * 2
            for block_layer, block in enumerate(block_list):
                if block_layer >= 2:
                    break
                if block in block_palette:
                    indexed_block[block_layer] = block_palette.index(block)
                else:
                    indexed_block[block_layer] = len(block_palette)
                    block_palette.append(block)
            block_palette_indices.append(indexed_block)

        block_indices = numpy.array(block_palette_indices, dtype=numpy.int32)[
            self._blocks
        ].T

        self._data["structure"] = amulet_nbt.TAG_Compound(
            {
                "block_indices": amulet_nbt.TAG_List(
                    [  # a list of tag ints that index into the block_palette. One list per block layer
                        amulet_nbt.TAG_List(
                            [amulet_nbt.TAG_Int(block) for block in layer]
                        )
                        for layer in block_indices
                    ]
                ),
                "entities": amulet_nbt.TAG_List(self._entities),
                "palette": amulet_nbt.TAG_Compound(
                    {
                        "default": amulet_nbt.TAG_Compound(
                            {
                                "block_palette": amulet_nbt.TAG_List(block_palette),
                                "block_position_data": amulet_nbt.TAG_Compound(
                                    {
                                        str(
                                            (
                                                (
                                                    block_entity["x"].value
                                                    - self._selection.min_x
                                                )
                                                * self._selection.size_y
                                                + (
                                                    block_entity["y"].value
                                                    - self._selection.min_y
                                                )
                                            )
                                            * self._selection.size_z
                                            + block_entity["z"].value
                                            - self._selection.min_z
                                        ): amulet_nbt.TAG_Compound(
                                            {"block_entity_data": block_entity}
                                        )
                                        for block_entity in self._block_entities
                                    }
                                ),
                            }
                        )
                    }
                ),
            }
        )
        self._data.save_to(self._path_or_buffer, compressed=False, little_endian=True)



from __future__ import annotations

import os
from amulet.world_interface.loader import Loader

SUPPORTED_INTERFACE_VERSION = 0
SUPPORTED_META_VERSION = 0

INTERFACES_DIRECTORY = os.path.dirname(__file__)

loader = Loader(
    "interface",
    INTERFACES_DIRECTORY,
    SUPPORTED_META_VERSION,
    SUPPORTED_INTERFACE_VERSION,
)

if __name__ == "__main__":
    import time

    print(loader.get_all())
    time.sleep(1)
    loader.reload()
    print(loader.get_all())



from __future__ import annotations

import os
import struct
from typing import (
    Tuple,
    Dict,
    Generator,
    Set,
    Union,
    Optional,
    List,
    TYPE_CHECKING,
    BinaryIO,
)
from io import BytesIO

import amulet_nbt as nbt

from amulet.utils.format_utils import check_all_exist
from amulet.api.errors import ChunkDoesNotExist, LevelDoesNotExist
from amulet.api.data_types import ChunkCoordinates
from amulet.api.wrapper.world_format_wrapper import WorldFormatWrapper
from amulet.world_interface.chunk import interfaces
from amulet.libs.leveldb import LevelDB

from amulet.world_interface.chunk.interfaces.leveldb.leveldb_chunk_versions import (
    game_to_chunk_version,
)

if TYPE_CHECKING:
    from amulet.api.wrapper import Interface
    from amulet.api.data_types import Dimension

InternalDimension = Optional[int]


class LevelDBLevelManager:
    # tag_ids = {45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 118}

    def __init__(self, directory: str):
        self._directory = directory
        self._db = LevelDB(os.path.join(self._directory, "db"))
        # self._levels format Dict[level, Dict[Tuple[cx, cz], List[Tuple[full_key, key_extension]]]]
        self._levels: Dict[InternalDimension, Set[ChunkCoordinates]] = {}
        self._dimension_name_map: Dict["Dimension", InternalDimension] = {}
        self._batch_temp: Dict[bytes, Union[bytes, None]] = {}

        self.register_dimension(None, "overworld")
        self.register_dimension(1, "nether")
        self.register_dimension(2, "end")

        for key in self._db.keys():
            if 9 <= len(key) <= 10 and key[8] in [44, 118]:  # "," "v"
                self._add_chunk(key)

            elif 13 <= len(key) <= 14 and key[12] in [44, 118]:  # "," "v"
                self._add_chunk(key, has_level=True)

    def save(self):
        batch = {}
        for key, val in self._batch_temp.items():
            if val is None:
                self._db.delete(key)
            else:
                batch[key] = val
        self._db.putBatch(batch)
        self._batch_temp.clear()

    def close(self):
        self._db.close()

    @property
    def dimensions(self) -> List["Dimension"]:
        """A list of all the levels contained in the world"""
        return list(self._dimension_name_map.keys())

    def register_dimension(
        self,
        dimension_internal: InternalDimension,
        dimension_name: Optional["Dimension"] = None,
    ):
        """
        Register a new dimension.
        :param dimension_internal: The internal representation of the dimension
        :param dimension_name: The name of the dimension shown to the user
        :return:
        """
        if dimension_name is None:
            dimension_name: "Dimension" = f"DIM{dimension_internal}"

        if (
            dimension_internal not in self._levels
            and dimension_name not in self._dimension_name_map
        ):
            self._levels[dimension_internal] = set()
            self._dimension_name_map[dimension_name] = dimension_internal

    def _get_internal_dimension(self, dimension: "Dimension") -> InternalDimension:
        if dimension in self._dimension_name_map:
            return self._dimension_name_map[dimension]
        else:
            raise LevelDoesNotExist(dimension)

    def all_chunk_coords(self, dimension: "Dimension") -> Set[ChunkCoordinates]:
        internal_dimension = self._get_internal_dimension(dimension)
        if internal_dimension in self._levels:
            return self._levels[internal_dimension]
        else:
            return set()

    def _add_chunk(self, key_: bytes, has_level: bool = False):
        if has_level:
            cx, cz, level = struct.unpack("<iii", key_[:12])
        else:
            cx, cz = struct.unpack("<ii", key_[:8])
            level = None
        if level not in self._levels:
            self.register_dimension(level)
        self._levels[level].add((cx, cz))

    def get_chunk_data(
        self, cx: int, cz: int, dimension: "Dimension"
    ) -> Dict[bytes, bytes]:
        """Get a dictionary of chunk key extension in bytes to the raw data in the key.
        chunk key extension are the character(s) after <cx><cz>[level] in the key
        Will raise ChunkDoesNotExist if the chunk does not exist
        """
        iter_start = struct.pack("<ii", cx, cz)
        iter_end = iter_start + b"\xff"
        internal_dimension = self._get_internal_dimension(dimension)
        if (
            internal_dimension in self._levels
            and (cx, cz) in self._levels[internal_dimension]
        ):
            chunk_data = {}
            for key, val in self._db.iterate(iter_start, iter_end):
                if internal_dimension is None:
                    if 9 <= len(key) <= 10:
                        key_extension = key[8:]
                    else:
                        continue
                else:
                    if (
                        13 <= len(key) <= 14
                        and struct.unpack("<i", key[8:12])[0] == internal_dimension
                    ):
                        key_extension = key[12:]
                    else:
                        continue
                chunk_data[key_extension] = val
            return chunk_data
        else:
            raise ChunkDoesNotExist

    def put_chunk_data(
        self, cx: int, cz: int, data: Dict[bytes, bytes], dimension: "Dimension"
    ):
        """pass data to the region file class"""
        # get the region key
        internal_dimension = self._get_internal_dimension(dimension)
        self._levels[internal_dimension].add((cx, cz))
        if internal_dimension is None:
            key_prefix = struct.pack("<ii", cx, cz)
        else:
            key_prefix = struct.pack("<iii", cx, cz, internal_dimension)
        for key, val in data.items():
            self._batch_temp[key_prefix + key] = val

    def delete_chunk(self, cx: int, cz: int, dimension: "Dimension"):
        if dimension in self._dimension_name_map:
            internal_dimension = self._dimension_name_map[dimension]
            if (
                internal_dimension in self._levels
                and (cx, cz) in self._levels[internal_dimension]
            ):
                chunk_data = self.get_chunk_data(cx, cz, dimension)
                self._levels[internal_dimension].remove((cx, cz))
                for key in chunk_data.keys():
                    if internal_dimension is None:
                        key_prefix = struct.pack("<ii", cx, cz)
                    else:
                        key_prefix = struct.pack("<iii", cx, cz, internal_dimension)

                    self._batch_temp[key_prefix + key] = None


class LevelDAT(nbt.NBTFile):
    def __init__(self, path: str):
        with open(path, "rb") as f:
            self._level_dat_version = struct.unpack("<i", f.read(4))[
                0
            ]  # TODO: handle other versions
            assert (
                4 <= self._level_dat_version <= 8
            ), f"Unknown level.dat version {self._level_dat_version}"
            data_length = struct.unpack("<i", f.read(4))[0]
            root_tag = nbt.load(
                buffer=f.read(data_length), compressed=False, little_endian=True
            )
        super().__init__(root_tag.value, root_tag.name)

    def save_to(
        self, filename_or_buffer: Union[str, BinaryIO] = None
    ) -> Optional[bytes]:
        buffer = BytesIO()
        buffer.write(struct.pack("<i", self._level_dat_version))
        buffer.write(super().save_to(compressed=False, little_endian=True))
        if filename_or_buffer is None:
            return buffer.getvalue()
        elif isinstance(filename_or_buffer, str):
            with open(filename_or_buffer, "wb") as f:
                f.write(buffer.getvalue())
        else:
            filename_or_buffer.write(buffer.getvalue())


class LevelDBFormat(WorldFormatWrapper):
    def __init__(self, directory: str):
        super().__init__(directory)
        self._lock = False
        self.root_tag: nbt.NBTFile = nbt.NBTFile()
        self._load_level_dat()
        self._level_manager: Optional[LevelDBLevelManager] = None

    def _load_level_dat(self):
        """Load the level.dat file and check the image file"""
        self.root_tag = LevelDAT(os.path.join(self.path, "level.dat"))
        if os.path.isfile(os.path.join(self.path, "world_icon.jpeg")):
            self._world_image_path = os.path.join(self.path, "world_icon.jpeg")

    @staticmethod
    def is_valid(directory):
        """
        Returns whether this format is able to load a given world.

        :param directory: The path to the root of the world to load.
        :return: True if the world can be loaded by this format, False otherwise.
        """
        if not check_all_exist(directory, "db", "level.dat", "levelname.txt"):
            return False

        return True

    @property
    def platform(self) -> str:
        """Platform string"""
        return "bedrock"

    def _get_version(self) -> Tuple[int, ...]:
        """The version the world was last opened in
        This should be greater than or equal to the chunk versions found within
        For this format wrapper it returns a tuple of 3/4 ints (the game version number)"""
        try:
            return tuple([t.value for t in self.root_tag["lastOpenedWithVersion"]])
        except Exception:
            return 1, 2, 0

    @property
    def world_name(self):
        """The name of the world"""
        return self.root_tag["LevelName"].value

    @world_name.setter
    def world_name(self, value: str):
        self.root_tag["LevelName"] = nbt.TAG_String(value)

    @property
    def last_played(self) -> int:
        return self.root_tag["LastPlayed"].value

    @property
    def game_version_string(self) -> str:
        try:
            return f'Bedrock {".".join(str(v.value) for v in self.root_tag["lastOpenedWithVersion"].value)}'
        except Exception:
            return f"Bedrock Unknown Version"

    @property
    def dimensions(self) -> List["Dimension"]:
        """A list of all the levels contained in the world"""
        self._verify_has_lock()
        return self._level_manager.dimensions

    def register_dimension(
        self, dimension_internal: int, dimension_name: Optional["Dimension"] = None
    ):
        """
        Register a new dimension.
        :param dimension_internal: The internal representation of the dimension
        :param dimension_name: The name of the dimension shown to the user
        :return:
        """
        self._level_manager.register_dimension(dimension_internal, dimension_name)

    def _get_interface(self, max_world_version, raw_chunk_data=None) -> "Interface":
        if raw_chunk_data:
            key = self._get_interface_key(raw_chunk_data)
        else:
            key = "bedrock", game_to_chunk_version(max_world_version[1])
        return interfaces.loader.get(key)

    def _get_interface_key(self, raw_chunk_data: Dict[bytes, bytes]) -> Tuple[str, int]:
        if b"," in raw_chunk_data:
            v = raw_chunk_data[b","][0]
        else:
            v = raw_chunk_data.get(b"v", "\x00")[0]
        return (
            self.platform,
            v,
        )  # TODO: work out a valid default

    def _reload_world(self):
        try:
            self.close()
        except:
            pass
        self._level_manager = LevelDBLevelManager(self.path)
        self._lock = True

    def open(self):
        """Open the database for reading and writing"""
        self._reload_world()

    @property
    def has_lock(self) -> bool:
        """Verify that the world database can be read and written"""
        # TODO: work out how to do this properly
        return self._lock

    def save(self):
        self._verify_has_lock()
        self._level_manager.save()
        self._changed = False

    def close(self):
        self._verify_has_lock()
        self._level_manager.close()

    def unload(self):
        """Unload data stored in the Format class"""
        pass

    def all_chunk_coords(
        self, dimension: "Dimension"
    ) -> Generator[ChunkCoordinates, None, None]:
        self._verify_has_lock()
        yield from self._level_manager.all_chunk_coords(dimension)

    def delete_chunk(self, cx: int, cz: int, dimension: "Dimension"):
        self._verify_has_lock()
        self._level_manager.delete_chunk(cx, cz, dimension)

    def _put_raw_chunk_data(
        self, cx: int, cz: int, data: Dict[bytes, bytes], dimension: "Dimension", *args
    ):
        """
        Actually stores the data from the interface to disk.
        """
        self._verify_has_lock()
        return self._level_manager.put_chunk_data(cx, cz, data, dimension)

    def _get_raw_chunk_data(
        self, cx: int, cz: int, dimension: "Dimension", *args
    ) -> Dict[bytes, bytes]:
        """
        Return the interface key and data to interface with given chunk coordinates.

        :param cx: The x coordinate of the chunk.
        :param cz: The z coordinate of the chunk.
        :return: The interface key for the get_interface method and the data to interface with.
        """
        self._verify_has_lock()
        return self._level_manager.get_chunk_data(cx, cz, dimension)


FORMAT_CLASS = LevelDBFormat



from __future__ import annotations

import os
import numpy
from typing import List, Dict, Tuple, Generator

import amulet_nbt
from amulet.api.selection import SelectionBox
from amulet.api.data_types import PointCoordinates, ChunkCoordinates, PathOrBuffer
from amulet.api.errors import ChunkDoesNotExist

BlockArrayType = numpy.ndarray  # uint16 array
BlockDataArrayType = numpy.ndarray  # uint8 array


class SchematicChunk:
    def __init__(
        self,
        selection: SelectionBox,
        blocks: BlockArrayType,
        data: BlockDataArrayType,
        block_entities: List[amulet_nbt.TAG_Compound],
        entities: List[amulet_nbt.TAG_Compound],
    ):
        self.selection = selection
        assert isinstance(blocks, numpy.ndarray) and blocks.dtype == numpy.uint16
        assert isinstance(data, numpy.ndarray) and data.dtype == numpy.uint8
        self.blocks = blocks
        self.data = data
        self.block_entities = block_entities
        self.entities = entities


class SchematicReader:
    def __init__(self, path_or_buffer: PathOrBuffer):
        if isinstance(path_or_buffer, str):
            assert path_or_buffer.endswith(
                ".schematic"
            ), "File selected is not a .schematic file"
            assert os.path.isfile(
                path_or_buffer
            ), f"There is no schematic file at path {path_or_buffer}"
            schematic = amulet_nbt.load(path_or_buffer)
            assert not all(
                key in schematic for key in ("Version", "Data Version", "BlockData")
            ), "This file is not a legacy schematic file."
        else:
            assert hasattr(path_or_buffer, "read"), "Object does not have a read method"
            schematic = amulet_nbt.load(buffer=path_or_buffer)

        materials = schematic.get("Materials", amulet_nbt.TAG_String()).value
        if materials == "Alpha":
            self._platform = "java"
        elif materials == "Pocket":
            self._platform = "bedrock"
        else:
            raise Exception(
                f'"{materials}" is not a supported platform for a schematic file.'
            )
        self._chunks: Dict[
            ChunkCoordinates,
            Tuple[SelectionBox, BlockArrayType, BlockDataArrayType, list, list],
        ] = {}
        self._selection = SelectionBox(
            (0, 0, 0),
            (
                schematic["Width"].value,
                schematic["Height"].value,
                schematic["Length"].value,
            ),
        )
        entities: amulet_nbt.TAG_List = schematic.get("Entities", amulet_nbt.TAG_List())
        block_entities: amulet_nbt.TAG_List = schematic.get(
            "TileEntities", amulet_nbt.TAG_List()
        )
        blocks: numpy.ndarray = (
            schematic["Blocks"].value.astype(numpy.uint8).astype(numpy.uint16)
        )
        if "AddBlocks" in schematic:
            add_blocks = schematic["AddBlocks"]
            blocks = (
                blocks
                + (
                    numpy.concatenate([[(add_blocks & 0xF0) >> 4], [add_blocks & 0xF]])
                    .T.ravel()
                    .astype(numpy.uint16)
                    << 8
                )[: blocks.size]
            )
        max_point = self._selection.max
        temp_shape = (max_point[1], max_point[2], max_point[0])
        blocks = numpy.transpose(blocks.reshape(temp_shape), (2, 0, 1))  # YZX => XYZ
        data = numpy.transpose(
            schematic["Data"].value.reshape(temp_shape), (2, 0, 1)
        ).astype(numpy.uint8)
        for cx, cz in self._selection.chunk_locations():
            box = SelectionBox(
                (cx * 16, 0, cz * 16),
                (
                    min((cx + 1) * 16, self._selection.size_x),
                    self._selection.size_y,
                    min((cz + 1) * 16, self._selection.size_z),
                ),
            )
            self._chunks[(cx, cz)] = (box, blocks[box.slice], data[box.slice], [], [])
        for e in block_entities:
            if all(key in e for key in ("x", "y", "z")):
                x, y, z = e["x"].value, e["y"].value, e["z"].value
                if (x, y, z) in self._selection:
                    cx = x >> 4
                    cz = z >> 4
                    self._chunks[(cx, cz)][3].append(e)
        for e in entities:
            if "Pos" in e:
                pos: PointCoordinates = tuple(map(lambda t: t.value, e["Pos"].value))
                if pos in self._selection:
                    cx = int(pos[0]) >> 4
                    cz = int(pos[2]) >> 4
                    self._chunks[(cx, cz)][4].append(e)

    def read(self, cx: int, cz: int):
        if (cx, cz) in self._chunks:
            return SchematicChunk(*self._chunks[(cx, cz)])
        else:
            raise ChunkDoesNotExist

    @property
    def platform(self) -> str:
        return self._platform

    @property
    def selection(self) -> SelectionBox:
        return self._selection

    @property
    def chunk_coords(self) -> Generator[ChunkCoordinates, None, None]:
        yield from self._chunks.keys()

    def close(self):
        pass


class SchematicWriter:
    def __init__(self, path_or_buffer: str, platform: str, selection: SelectionBox):
        self._path_or_buffer = path_or_buffer
        if platform == "java":
            self._materials = "Alpha"
        elif platform == "bedrock":
            self._materials = "Pocket"
        else:
            raise Exception(
                f'"{platform}" is not a supported platform for a schematic file.'
            )
        self._selection = selection

        self._data = amulet_nbt.NBTFile(
            amulet_nbt.TAG_Compound(
                {
                    "TileTicks": amulet_nbt.TAG_List(),
                    "Width": amulet_nbt.TAG_Short(selection.size_x),
                    "Height": amulet_nbt.TAG_Short(selection.size_y),
                    "Length": amulet_nbt.TAG_Short(selection.size_z),
                    "Materials": amulet_nbt.TAG_String(self._materials),
                }
            ),
            "Schematic",
        )

        self._entities = []
        self._block_entities = []
        self._blocks = numpy.zeros(
            selection.shape, dtype=numpy.uint16
        )  # only 12 bits are actually used at most
        self._block_data = numpy.zeros(
            selection.shape, dtype=numpy.uint8
        )  # only 4 bits are used

    def write(self, section: SchematicChunk):
        if section.selection.intersects(self._selection):
            box = section.selection.create_moved_box(self._selection.min, subtract=True)
            self._blocks[box.slice] = section.blocks
            self._block_data[box.slice] = section.data
            for be in section.block_entities:
                coord_type = be["x"].__class__
                be["x"] = coord_type(be["x"] - self._selection.min_x)
                be["y"] = coord_type(be["y"] - self._selection.min_y)
                be["z"] = coord_type(be["z"] - self._selection.min_z)
                self._block_entities.append(be)
            for e in section.entities:
                coord_type = e["Pos"][0].__class__
                e["Pos"][0] = coord_type(e["Pos"][0] - self._selection.min_x)
                e["Pos"][1] = coord_type(e["Pos"][1] - self._selection.min_y)
                e["Pos"][2] = coord_type(e["Pos"][2] - self._selection.min_z)
                self._entities.append(e)

    def close(self):
        self._data["Entities"] = amulet_nbt.TAG_List(self._entities)
        self._data["TileEntities"] = amulet_nbt.TAG_List(self._block_entities)
        self._data["Data"] = amulet_nbt.TAG_Byte_Array(
            numpy.transpose(self._block_data, (1, 2, 0))  # XYZ => YZX
        )
        self._data["Blocks"] = amulet_nbt.TAG_Byte_Array(
            numpy.transpose((self._blocks & 0xFF).astype(numpy.uint8), (1, 2, 0))
        )
        if numpy.max(self._blocks) > 0xFF:
            add_blocks = numpy.transpose(self._blocks & 0xF00, (1, 2, 0)) >> 8
            self._data["AddBlocks"] = amulet_nbt.TAG_Byte_Array(
                add_blocks[::2] + (add_blocks[1::2] << 4)
            )
        self._data.save_to(self._path_or_buffer)



from __future__ import annotations

import glob
import importlib.util
import json
import os
from typing import AbstractSet, Any, Dict

from amulet import log
from amulet.api.errors import LoaderNoneMatched


class Loader:
    def __init__(
        self,
        object_type: str,
        directory: str,
        supported_meta_version,
        supported_version,
        create_instance=True,
    ):
        self._object_type = object_type
        self._directory = directory
        self._supported_meta_version = supported_meta_version
        self._supported_version = supported_version
        self._loaded: Dict[str, Any] = {}
        self._is_loaded = False
        self._create_instance = create_instance

    def _find(self):
        """Load all objects from the object directory"""

        directories = glob.iglob(
            os.path.join(glob.escape(self._directory), "**", ""), recursive=True
        )
        for d in directories:
            meta_path = os.path.join(d, f"{self._object_type}.meta")
            if not os.path.exists(meta_path):
                continue

            with open(meta_path) as fp:
                meta = json.load(fp)

            if meta["meta_version"] != self._supported_meta_version:
                log.error(
                    f'Could not enable {self._object_type} located in "{d}" due to unsupported meta version'
                )
                continue

            if (
                meta[self._object_type][f"{self._object_type}_version"]
                != self._supported_version
            ):
                log.error(
                    f'Could not enable {self._object_type} "{meta[self._object_type]["id"]}" due to unsupported {self._object_type} version'
                )
                continue

            spec = importlib.util.spec_from_file_location(
                meta[self._object_type]["entry_point"],
                os.path.join(d, meta[self._object_type]["entry_point"] + ".py"),
            )
            modu = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(modu)

            if not hasattr(modu, f"{self._object_type.upper()}_CLASS"):
                log.error(
                    f'{self._object_type} "{meta[self._object_type]["id"]}" is missing the {self._object_type.upper()}_CLASS attribute'
                )
                continue

            c = getattr(modu, f"{self._object_type.upper()}_CLASS")
            if self._create_instance:
                self._loaded[meta[f"{self._object_type}"]["id"]] = c()
            else:
                self._loaded[meta[f"{self._object_type}"]["id"]] = c

            log.debug(
                f'Enabled {self._object_type} "{meta[self._object_type]["id"]}", version {meta[self._object_type]["wrapper_version"]}'
            )

        self._is_loaded = True

    def reload(self):
        """Reloads all objects"""
        self._loaded.clear()
        self._find()

    def get_all(self) -> AbstractSet[str]:
        """
        :return: The identifiers of all loaded objects
        """
        self._load()
        return self._loaded.keys()

    def get(self, identifier: Any) -> Any:
        """
        Given an ``identifier`` will find a valid class and return it

        :param identifier: The identifier for the desired loaded object
        :return: The class for the object
        """
        object_id = self.identify(identifier)
        return self._loaded[object_id]

    def identify(self, identifier: Any) -> str:

        self._load()

        for object_name, object_instance in self._loaded.items():
            if object_instance.is_valid(identifier):
                return object_name

        raise LoaderNoneMatched(
            f"Could not find a matching {self._object_type} for {identifier}"
        )

    def get_by_id(self, object_id: str):
        return self._loaded[object_id]

    def _load(self):
        if not self._is_loaded:
            self._find()

    def __contains__(self, item: str):
        self._load()
        return item in self._loaded



from __future__ import annotations

from amulet.world_interface.chunk.interfaces.leveldb.leveldb_10.interface import (
    LevelDB10Interface,
)


class LevelDB11Interface(LevelDB10Interface):
    def __init__(self):
        super().__init__()

        self.features["chunk_version"] = 11


INTERFACE_CLASS = LevelDB11Interface



from __future__ import annotations

from typing import Sequence, Callable, TypeVar, Generic, Optional

T = TypeVar("T")


class Stack(Generic[T]):
    """
    Implementation of a Stack, but prevents out-of-order editing of stored data
    """

    def __init__(self, initial_data: Sequence[T] = None):
        self._data = initial_data or []
        self.__contains__ = self._data.__contains__
        self.pop: Callable[[], T] = self._data.pop
        self.append: Callable[[T], None] = self._data.append
        self.size: Callable[[], int] = self._data.__len__

    def peek(self) -> Optional[T]:
        """
        Returns the top element of the Stack without removing it

        :return: The top element of the Stack, None if the Stack is empty
        """
        if self.is_empty():
            return None

        return self._data[-1]

    def is_empty(self) -> bool:
        """
        Checks if the Stack is empty

        :return: True is the Stack holds elements, False otherwise
        """
        return len(self._data) == 0

    def clear(self):
        """
        Clears all stored elements in the Stack
        """
        self._data.clear()



from __future__ import annotations

import itertools
import numpy

from typing import Tuple, Iterable, List, Generator, Dict

from amulet.api.data_types import (
    BlockCoordinatesAny,
    CoordinatesAny,
    ChunkCoordinates,
    SubChunkCoordinates,
    FloatTriplet,
)
from ..utils.world_utils import (
    block_coords_to_chunk_coords,
    blocks_slice_to_chunk_slice,
)
from amulet.utils.matrix import transform_matrix


class SelectionBox:
    """
    A SelectionBox is a box that can represent the entirety of a selection or just a subsection
    of one. This allows for non-rectangular and non-contiguous selections.

    The both the minimum and  maximum coordinate points are inclusive.
    """

    def __init__(self, min_point: BlockCoordinatesAny, max_point: BlockCoordinatesAny):
        box = numpy.array([min_point, max_point]).round().astype(numpy.int)
        self._min_x, self._min_y, self._min_z = numpy.min(box, 0).tolist()
        self._max_x, self._max_y, self._max_z = numpy.max(box, 0).tolist()

    @classmethod
    def create_chunk_box(cls, cx: int, cz: int, chunk_size: int = 16):
        """Get a SelectionBox containing the whole of a given chunk"""
        return cls(
            (cx * chunk_size, -(2 ** 30), cz * chunk_size),
            ((cx + 1) * chunk_size, 2 ** 30, (cz + 1) * chunk_size),
        )

    @classmethod
    def create_sub_chunk_box(cls, cx: int, cy: int, cz: int, chunk_size: int = 16):
        """Get a SelectionBox containing the whole of a given sub-chunk"""
        return cls(
            (cx * chunk_size, cy * chunk_size, cz * chunk_size),
            ((cx + 1) * chunk_size, (cy + 1) * chunk_size, (cz + 1) * chunk_size),
        )

    def create_moved_box(
        self, offset: BlockCoordinatesAny, subtract=False
    ) -> SelectionBox:
        """Create a new SelectionBox by offsetting the bounds of this box."""
        offset = numpy.array(offset)
        if subtract:
            offset *= -1
        return SelectionBox(offset + self.min, offset + self.max)

    def chunk_locations(
        self, chunk_size: int = 16
    ) -> Generator[ChunkCoordinates, None, None]:
        """A generator of chunk locations that this box intersects."""
        cx_min, cz_min, cx_max, cz_max = block_coords_to_chunk_coords(
            self.min_x,
            self.min_z,
            self.max_x - 1,
            self.max_z - 1,
            chunk_size=chunk_size,
        )
        yield from itertools.product(
            range(cx_min, cx_max + 1), range(cz_min, cz_max + 1)
        )

    def chunk_y_locations(self, chunk_size: int = 16):
        cy_min, cy_max = block_coords_to_chunk_coords(
            self.min_y, self._max_y, chunk_size=chunk_size
        )
        for cy in range(cy_min, cy_max + 1):
            yield cy

    def sub_chunk_locations(
        self, chunk_size: int = 16
    ) -> Generator[SubChunkCoordinates, None, None]:
        for cx, cz in self.chunk_locations(chunk_size):
            for cy in self.chunk_y_locations(chunk_size):
                yield cx, cy, cz

    def sub_sections(
        self, chunk_size: int = 16
    ) -> Generator[Tuple[ChunkCoordinates, SelectionBox], None, None]:
        """A generator of modified `SelectionBox`es to fit within each sub-chunk.
        :param chunk_size: The dimension of the chunk (normally 16)
        """
        for cx, cy, cz in self.sub_chunk_locations(chunk_size):
            yield (cx, cz), self.intersection(
                SelectionBox.create_sub_chunk_box(cx, cy, cz, chunk_size)
            )

    def __iter__(self) -> Iterable[Tuple[int, int, int]]:
        return self.blocks()

    def blocks(self) -> Iterable[Tuple[int, int, int]]:
        return itertools.product(
            range(self._min_x, self._max_x),
            range(self._min_y, self._max_y),
            range(self._min_z, self._max_z),
        )

    def __str__(self):
        return f"({self.min}, {self.max})"

    def __contains__(self, item: CoordinatesAny):
        return (
            self._min_x <= item[0] <= self._max_x
            and self._min_y <= item[1] <= self._max_y
            and self._min_z <= item[2] <= self._max_z
        )

    def __eq__(self, other):
        return self.min == other.min and self.max == other.max

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash((*self.min, *self.max))

    @property
    def slice(self) -> Tuple[slice, slice, slice]:
        """
        Converts the SubSelectionBoxes minimum/maximum coordinates into slice arguments

        :return: The SubSelectionBoxes coordinates as slices in (x,y,z) order
        """
        return (
            slice(self._min_x, self._max_x),
            slice(self._min_y, self._max_y),
            slice(self._min_z, self._max_z),
        )

    def chunk_slice(
        self, cx: int, cz: int, chunk_size: int = 16
    ) -> Tuple[slice, slice, slice]:
        """Get the Convert a slice in absolute coordinates to chunk coordinates"""
        s_x, s_y, s_z = self.slice
        x_chunk_slice = blocks_slice_to_chunk_slice(s_x, chunk_size, cx)
        z_chunk_slice = blocks_slice_to_chunk_slice(s_z, chunk_size, cz)
        return x_chunk_slice, s_y, z_chunk_slice

    @property
    def min_x(self) -> int:
        """The minimum x coordinate"""
        return self._min_x

    @property
    def min_y(self) -> int:
        """The minimum y coordinate"""
        return self._min_y

    @property
    def min_z(self) -> int:
        """The minimum z coordinate"""
        return self._min_z

    @property
    def max_x(self) -> int:
        """The maximum x coordinate"""
        return self._max_x

    @property
    def max_y(self) -> int:
        """The maximum y coordinate"""
        return self._max_y

    @property
    def max_z(self) -> int:
        """The maximum z coordinate"""
        return self._max_z

    @property
    def min(self) -> Tuple[int, int, int]:
        """The minimum point of the box"""
        return self._min_x, self._min_y, self._min_z

    @property
    def max(self) -> Tuple[int, int, int]:
        """The maximum point of the box"""
        return self._max_x, self._max_y, self._max_z

    @property
    def bounds(self) -> Tuple[int, int, int, int, int, int]:
        return (
            self._min_x,
            self._min_y,
            self._min_z,
            self._max_x,
            self._max_y,
            self._max_z,
        )

    @property
    def size_x(self) -> int:
        """The length of the box in the x axis"""
        return self._max_x - self._min_x

    @property
    def size_y(self) -> int:
        """The length of the box in the y axis"""
        return self._max_y - self._min_y

    @property
    def size_z(self) -> int:
        """The length of the box in the z axis"""
        return self._max_z - self._min_z

    @property
    def shape(self) -> Tuple[int, int, int]:
        """The shape of the box"""
        return self.size_x, self.size_y, self.size_z

    @property
    def volume(self) -> int:
        return self.size_x * self.size_y * self.size_z

    def intersects(self, other: SelectionBox) -> bool:
        """
        Method to check whether this instance of SelectionBox intersects another SelectionBox

        :param other: The other SelectionBox to check for intersection
        :return: True if the two SubSelectionBoxes intersect, False otherwise
        """
        return not (
            self.min_x >= other.max_x
            or self.min_y >= other.max_y
            or self.min_z >= other.max_z
            or self.max_x <= other.min_x
            or self.max_y <= other.min_y
            or self.max_z <= other.min_z
        )

    def intersection(self, other: SelectionBox) -> SelectionBox:
        """Get a SelectionBox that represents the region contained within self and other.
        Box may be a zero width box. Use self.intersects to check that it actually intersects."""
        return SelectionBox(
            numpy.clip(other.min, self.min, self.max),
            numpy.clip(other.max, self.min, self.max),
        )

    def transform(
        self, scale: FloatTriplet, rotation: FloatTriplet
    ) -> List[SelectionBox]:
        """creates a list of new transformed SelectionBox(es)."""
        boxes = []
        # TODO: allow this to support rotations that are not 90 degrees
        min_point, max_point = numpy.matmul(
            transform_matrix((0, 0, 0), scale, rotation, "zyx"),
            numpy.array([[*self.min, 1], [*self.max, 1]]).T,
        ).T[:, :3]
        boxes.append(SelectionBox(min_point, max_point))

        return boxes


class SelectionGroup:
    """
    Holding class for multiple SubSelectionBoxes which allows for non-rectangular and non-contiguous selections
    """

    def __init__(self, selection_boxes: Iterable[SelectionBox] = ()):
        self._selection_boxes = []

        if selection_boxes:
            for box in selection_boxes:
                self.add_box(box)

    def __iter__(self) -> Iterable[Tuple[int, int, int]]:
        return itertools.chain.from_iterable(sorted(self._selection_boxes, key=hash))

    def __len__(self):
        return len(self._selection_boxes)

    def __contains__(self, item: CoordinatesAny):
        for subbox in self._selection_boxes:
            if item in subbox:
                return True

        return False

    def __bool__(self):
        return bool(self._selection_boxes)

    @property
    def min(self) -> numpy.ndarray:
        if self._selection_boxes:
            return numpy.min(numpy.array([box.min for box in self._selection_boxes]), 0)
        else:
            raise ValueError("SelectionGroup does not contain any SubSelectionBoxes")

    @property
    def max(self) -> numpy.ndarray:
        if self._selection_boxes:
            return numpy.max(numpy.array([box.max for box in self._selection_boxes]), 0)
        else:
            raise ValueError("SelectionGroup does not contain any SubSelectionBoxes")

    def add_box(self, other: SelectionBox, do_merge_check: bool = True):
        """
        Adds a SelectionBox to the selection box. If `other` is next to another SelectionBox in the selection, matches in any 2 dimensions, and
        `do_merge_check` is True, then the 2 boxes will be combined into 1 box.

        :param other: The box to add
        :param do_merge_check: Boolean flag to merge boxes if able
        """
        # TODO: verify that this logic actually works on more complex cases
        if do_merge_check:
            boxes_to_remove = None
            new_box = None
            for box in self._selection_boxes:
                x_dim = box.min_x == other.min_x and box.max_x == other.max_x
                y_dim = box.min_y == other.min_y and box.max_y == other.max_y
                z_dim = box.min_z == other.min_z and box.max_z == other.max_z

                x_border = box.max_x == other.min_x or other.max_x == box.min_x
                y_border = box.max_y == other.min_y or other.max_y == box.min_y
                z_border = box.max_z == other.min_z or other.max_z == box.min_z

                if (
                    (x_dim and y_dim and z_border)
                    or (x_dim and z_dim and y_border)
                    or (y_dim and z_dim and x_border)
                ):
                    boxes_to_remove = box
                    new_box = SelectionBox(box.min, other.max)
                    break

            if new_box:
                self._selection_boxes.remove(boxes_to_remove)
                self.add_box(new_box)
            else:
                self._selection_boxes.append(other)
        else:
            self._selection_boxes.append(other)

    @property
    def is_contiguous(self) -> bool:
        """Does the SelectionGroup represent one connected region (True) or multiple separated regions (False)"""
        if len(self._selection_boxes) == 1:
            return True

        for i in range(len(self._selection_boxes) - 1):
            sub_box = self._selection_boxes[i]
            next_box = self._selection_boxes[i + 1]
            if (
                abs(sub_box.max_x - next_box.min_x)
                and abs(sub_box.max_y - next_box.min_y)
                and abs(sub_box.max_z - next_box.min_z)
            ):
                return False

        return True

    @property
    def is_rectangular(self) -> bool:
        """
        Checks if the SelectionGroup is a rectangle

        :return: True is the selection is a rectangle, False otherwise
        """
        return len(self._selection_boxes) == 1

    @property
    def selection_boxes(self) -> List[SelectionBox]:
        """
        Returns a list of unmodified SubSelectionBoxes in the SelectionGroup.
        :return: A list of the SubSelectionBoxes
        """
        return sorted(self._selection_boxes.copy(), key=hash)

    def chunk_locations(
        self, chunk_size: int = 16
    ) -> Generator[ChunkCoordinates, None, None]:
        """The chunk locations that the SelectionGroup is in.
        Each location is only given once even if there are multiple boxes in the chunk."""
        yield from set(
            location
            for box in self.selection_boxes
            for location in box.chunk_locations(chunk_size)
        )

    def _chunk_boxes(
        self, chunk_size: int = 16
    ) -> Dict[ChunkCoordinates, List[SelectionBox]]:
        boxes = {}
        for box in self.selection_boxes:
            for (cx, cz), sub_box in box.sub_sections(chunk_size):
                boxes.setdefault((cx, cz), []).append(sub_box)
        return boxes

    def sub_sections(
        self, chunk_size: int = 16
    ) -> Generator[Tuple[ChunkCoordinates, SelectionBox], None, None]:
        """A generator of modified `SelectionBox`es to fit within each sub-chunk.
        :param chunk_size: The dimension of the chunk (normally 16)
        """
        for (cx, cz), boxes in self._chunk_boxes(chunk_size).items():
            for box in boxes:
                yield (cx, cz), box

    def sub_slices(
        self, chunk_size: int = 16
    ) -> Generator[
        Tuple[ChunkCoordinates, Tuple[slice, slice, slice], SelectionBox], None, None
    ]:
        for (cx, cz), box in self.sub_sections(chunk_size):
            slices = box.chunk_slice(cx, cz, chunk_size)
            yield (cx, cz), slices, box

    def intersects(self, other: SelectionGroup) -> bool:
        """Check if self and other intersect"""
        return any(
            self_box.intersects(other_box)
            for self_box in self.selection_boxes
            for other_box in other.selection_boxes
        )

    def intersection(self, other: SelectionGroup) -> SelectionGroup:
        """Get a new SelectionGroup that represents the area contained within self and other"""
        intersection = SelectionGroup()
        for self_box in self.selection_boxes:
            for other_box in other.selection_boxes:
                if self_box.intersects(other_box):
                    intersection.add_box(self_box.intersection(other_box))
        return intersection

    def transform(self, scale: FloatTriplet, rotation: FloatTriplet) -> SelectionGroup:
        """creates a new transformed SelectionGroup."""
        selection_group = SelectionGroup()
        for selection in self.selection_boxes:
            for transformed_selection in selection.transform(scale, rotation):
                selection_group.add_box(transformed_selection)
        return selection_group


if __name__ == "__main__":
    b1 = SelectionBox((0, 0, 0), (4, 4, 4))
    b2 = SelectionBox((7, 7, 7), (10, 10, 10))
    sel_box = SelectionGroup((b1, b2))

    for x, y, z in sel_box:
        print(x, y, z)



from __future__ import annotations

from amulet.world_interface.chunk.interfaces.leveldb.leveldb_12.interface import (
    LevelDB12Interface,
)


class LevelDB13Interface(LevelDB12Interface):
    def __init__(self):
        super().__init__()

        self.features["chunk_version"] = 13


INTERFACE_CLASS = LevelDB13Interface



# Provides an python wrapper for the ctypes wrapper for the c wrapper for leveldb.

import ctypes
import os.path as op
import sys

if sys.platform == "win32":
    if sys.maxsize > 2 ** 32:  # 64 bit python
        ldb = ctypes.cdll.LoadLibrary(
            op.join(op.dirname(op.realpath(__file__)), "LevelDB-MCPE-64.dll")
        )
    else:  # 32 bit python
        ldb = ctypes.cdll.LoadLibrary(
            op.join(op.dirname(op.realpath(__file__)), "LevelDB-MCPE-32.dll")
        )
elif sys.platform == "darwin":
    ldb = ctypes.cdll.LoadLibrary(
        op.join(op.dirname(op.realpath(__file__)), "libleveldb.dylib")
    )
else:  # linux, compile your own .so if this errors!
    ldb = ctypes.cdll.LoadLibrary(
        op.join(op.dirname(op.realpath(__file__)), "libleveldb.so")
    )  # Load DLL

# Setup ctypes arguments and return types for all of the leveldb functions.
# Most of this pulled from Podshot/MCEdit-Unified
ldb.leveldb_filterpolicy_create_bloom.argtypes = [ctypes.c_int]
ldb.leveldb_filterpolicy_create_bloom.restype = ctypes.c_void_p
ldb.leveldb_filterpolicy_destroy.argtypes = [ctypes.c_void_p]
ldb.leveldb_filterpolicy_destroy.restype = None
ldb.leveldb_cache_create_lru.argtypes = [ctypes.c_size_t]
ldb.leveldb_cache_create_lru.restype = ctypes.c_void_p
ldb.leveldb_cache_destroy.argtypes = [ctypes.c_void_p]
ldb.leveldb_cache_destroy.restype = None

ldb.leveldb_options_create.argtypes = []
ldb.leveldb_options_create.restype = ctypes.c_void_p
ldb.leveldb_options_set_filter_policy.argtypes = [ctypes.c_void_p, ctypes.c_void_p]
ldb.leveldb_options_set_filter_policy.restype = None
ldb.leveldb_options_set_create_if_missing.argtypes = [ctypes.c_void_p, ctypes.c_ubyte]
ldb.leveldb_options_set_create_if_missing.restype = None
ldb.leveldb_options_set_error_if_exists.argtypes = [ctypes.c_void_p, ctypes.c_ubyte]
ldb.leveldb_options_set_error_if_exists.restype = None
ldb.leveldb_options_set_paranoid_checks.argtypes = [ctypes.c_void_p, ctypes.c_ubyte]
ldb.leveldb_options_set_paranoid_checks.restype = None
ldb.leveldb_options_set_write_buffer_size.argtypes = [ctypes.c_void_p, ctypes.c_size_t]
ldb.leveldb_options_set_write_buffer_size.restype = None
ldb.leveldb_options_set_max_open_files.argtypes = [ctypes.c_void_p, ctypes.c_int]
ldb.leveldb_options_set_max_open_files.restype = None
ldb.leveldb_options_set_cache.argtypes = [ctypes.c_void_p, ctypes.c_void_p]
ldb.leveldb_options_set_cache.restype = None
ldb.leveldb_options_set_block_size.argtypes = [ctypes.c_void_p, ctypes.c_size_t]
ldb.leveldb_options_set_block_size.restype = None
ldb.leveldb_options_destroy.argtypes = [ctypes.c_void_p]
ldb.leveldb_options_destroy.restype = None

ldb.leveldb_options_set_compression.argtypes = [ctypes.c_void_p, ctypes.c_int]
ldb.leveldb_options_set_compression.restype = None

ldb.leveldb_open.argtypes = [ctypes.c_void_p, ctypes.c_char_p, ctypes.c_void_p]
ldb.leveldb_open.restype = ctypes.c_void_p
ldb.leveldb_close.argtypes = [ctypes.c_void_p]
ldb.leveldb_close.restype = None
ldb.leveldb_put.argtypes = [
    ctypes.c_void_p,
    ctypes.c_void_p,
    ctypes.c_void_p,
    ctypes.c_size_t,
    ctypes.c_void_p,
    ctypes.c_size_t,
    ctypes.c_void_p,
]
ldb.leveldb_put.restype = None
ldb.leveldb_delete.argtypes = [
    ctypes.c_void_p,
    ctypes.c_void_p,
    ctypes.c_void_p,
    ctypes.c_size_t,
    ctypes.c_void_p,
]
ldb.leveldb_delete.restype = None
ldb.leveldb_write.argtypes = [
    ctypes.c_void_p,
    ctypes.c_void_p,
    ctypes.c_void_p,
    ctypes.c_void_p,
]
ldb.leveldb_write.restype = None
ldb.leveldb_get.argtypes = [
    ctypes.c_void_p,
    ctypes.c_void_p,
    ctypes.c_void_p,
    ctypes.c_size_t,
    ctypes.c_void_p,
    ctypes.c_void_p,
]
ldb.leveldb_get.restype = ctypes.POINTER(ctypes.c_char)

ldb.leveldb_writeoptions_create.argtypes = []
ldb.leveldb_writeoptions_create.restype = ctypes.c_void_p
ldb.leveldb_writeoptions_destroy.argtypes = [ctypes.c_void_p]
ldb.leveldb_writeoptions_destroy.restype = None
ldb.leveldb_writeoptions_set_sync.argtypes = [ctypes.c_void_p, ctypes.c_ubyte]
ldb.leveldb_writeoptions_set_sync.restype = None

ldb.leveldb_readoptions_create.argtypes = []
ldb.leveldb_readoptions_create.restype = ctypes.c_void_p
ldb.leveldb_readoptions_destroy.argtypes = [ctypes.c_void_p]
ldb.leveldb_readoptions_destroy.restype = None
ldb.leveldb_readoptions_set_verify_checksums.argtypes = [
    ctypes.c_void_p,
    ctypes.c_ubyte,
]
ldb.leveldb_readoptions_set_verify_checksums.restype = None
ldb.leveldb_readoptions_set_fill_cache.argtypes = [ctypes.c_void_p, ctypes.c_ubyte]
ldb.leveldb_readoptions_set_fill_cache.restype = None
ldb.leveldb_readoptions_set_snapshot.argtypes = [ctypes.c_void_p, ctypes.c_void_p]
ldb.leveldb_readoptions_set_snapshot.restype = None

ldb.leveldb_create_iterator.argtypes = [ctypes.c_void_p, ctypes.c_void_p]
ldb.leveldb_create_iterator.restype = ctypes.c_void_p
ldb.leveldb_iter_destroy.argtypes = [ctypes.c_void_p]
ldb.leveldb_iter_destroy.restype = None
ldb.leveldb_iter_valid.argtypes = [ctypes.c_void_p]
ldb.leveldb_iter_valid.restype = ctypes.c_bool
ldb.leveldb_iter_key.argtypes = [ctypes.c_void_p, ctypes.POINTER(ctypes.c_size_t)]
ldb.leveldb_iter_key.restype = ctypes.c_void_p
ldb.leveldb_iter_value.argtypes = [ctypes.c_void_p, ctypes.POINTER(ctypes.c_size_t)]
ldb.leveldb_iter_value.restype = ctypes.c_void_p
ldb.leveldb_iter_next.argtypes = [ctypes.c_void_p]
ldb.leveldb_iter_next.restype = None
ldb.leveldb_iter_prev.argtypes = [ctypes.c_void_p]
ldb.leveldb_iter_prev.restype = None
ldb.leveldb_iter_seek_to_first.argtypes = [ctypes.c_void_p]
ldb.leveldb_iter_seek_to_first.restype = None
ldb.leveldb_iter_seek_to_last.argtypes = [ctypes.c_void_p]
ldb.leveldb_iter_seek_to_last.restype = None
ldb.leveldb_iter_seek.argtypes = [ctypes.c_void_p, ctypes.c_void_p, ctypes.c_size_t]
ldb.leveldb_iter_seek.restype = None
ldb.leveldb_iter_get_error.argtypes = [ctypes.c_void_p, ctypes.c_void_p]
ldb.leveldb_iter_get_error.restype = None

ldb.leveldb_writebatch_create.argtypes = []
ldb.leveldb_writebatch_create.restype = ctypes.c_void_p
ldb.leveldb_writebatch_destroy.argtypes = [ctypes.c_void_p]
ldb.leveldb_writebatch_destroy.restype = None
ldb.leveldb_writebatch_clear.argtypes = [ctypes.c_void_p]
ldb.leveldb_writebatch_clear.restype = None

ldb.leveldb_writebatch_put.argtypes = [
    ctypes.c_void_p,
    ctypes.c_void_p,
    ctypes.c_size_t,
    ctypes.c_void_p,
    ctypes.c_size_t,
]
ldb.leveldb_writebatch_put.restype = None
ldb.leveldb_writebatch_delete.argtypes = [
    ctypes.c_void_p,
    ctypes.c_void_p,
    ctypes.c_size_t,
]
ldb.leveldb_writebatch_delete.restype = None

ldb.leveldb_approximate_sizes.argtypes = [
    ctypes.c_void_p,
    ctypes.c_int,
    ctypes.c_void_p,
    ctypes.c_void_p,
    ctypes.c_void_p,
    ctypes.c_void_p,
    ctypes.c_void_p,
]
ldb.leveldb_approximate_sizes.restype = None

ldb.leveldb_compact_range.argtypes = [
    ctypes.c_void_p,
    ctypes.c_void_p,
    ctypes.c_size_t,
    ctypes.c_void_p,
    ctypes.c_size_t,
]
ldb.leveldb_compact_range.restype = None

ldb.leveldb_create_snapshot.argtypes = [ctypes.c_void_p]
ldb.leveldb_create_snapshot.restype = ctypes.c_void_p
ldb.leveldb_release_snapshot.argtypes = [ctypes.c_void_p, ctypes.c_void_p]
ldb.leveldb_release_snapshot.restype = None

ldb.leveldb_free.argtypes = [ctypes.c_void_p]
ldb.leveldb_free.restype = None


class LevelDBException(Exception):
    pass


def _checkError(err):
    """Utility function for checking the error code returned by some leveldb functions."""
    if bool(err):  # Not an empty null-terminated string
        message = ctypes.string_at(err)
        ldb.leveldb_free(ctypes.cast(err, ctypes.c_void_p))
        raise LevelDBException(message)


class LevelDB:
    def __init__(self, path):
        self.db = None
        self._open(path)

    def __del__(self):
        self.close()

    def _open(self, path):
        # Bloom filter: an efficient way to tell if something is in a cache.
        filter_policy = ldb.leveldb_filterpolicy_create_bloom(10)
        cache = ldb.leveldb_cache_create_lru(40 * 1024 * 1024)
        options = ldb.leveldb_options_create()
        # Many of these options were pulled from Podshot/MCEdit-Unified
        ldb.leveldb_options_set_compression(options, 4)
        ldb.leveldb_options_set_filter_policy(options, filter_policy)
        ldb.leveldb_options_set_create_if_missing(options, False)
        ldb.leveldb_options_set_write_buffer_size(options, 4 * 1024 * 1024)
        ldb.leveldb_options_set_cache(options, cache)
        ldb.leveldb_options_set_block_size(options, 163840)

        error = ctypes.POINTER(ctypes.c_char)()
        db = ldb.leveldb_open(options, path.encode("utf-8"), ctypes.byref(error))
        ldb.leveldb_options_destroy(options)
        _checkError(error)

        self.db = db

    def close(self):
        if self.db:
            ldb.leveldb_compact_range(self.db, None, 0, None, 0)
            ldb.leveldb_close(self.db)
            self.db = None

    def get(self, key):
        ro = ldb.leveldb_readoptions_create()
        size = ctypes.c_size_t(0)
        error = ctypes.POINTER(ctypes.c_char)()
        valPtr = ldb.leveldb_get(
            self.db, ro, key, len(key), ctypes.byref(size), ctypes.byref(error)
        )
        ldb.leveldb_readoptions_destroy(ro)
        _checkError(error)
        if bool(valPtr):
            val = ctypes.string_at(valPtr, size.value)
            ldb.leveldb_free(ctypes.cast(valPtr, ctypes.c_void_p))
        else:
            raise KeyError("Key {} not found in database.".format(key))
        return val

    def put(self, key, val):
        wo = ldb.leveldb_writeoptions_create()
        error = ctypes.POINTER(ctypes.c_char)()
        ldb.leveldb_put(self.db, wo, key, len(key), val, len(val), ctypes.byref(error))
        ldb.leveldb_writeoptions_destroy(wo)
        _checkError(error)

    def putBatch(self, data):
        batch = ldb.leveldb_writebatch_create()
        for k, v in data.items():
            ldb.leveldb_writebatch_put(batch, k, len(k), v, len(v))
        wo = ldb.leveldb_writeoptions_create()
        error = ctypes.POINTER(ctypes.c_char)()
        ldb.leveldb_write(self.db, wo, batch, ctypes.byref(error))
        ldb.leveldb_writeoptions_destroy(wo)
        _checkError(error)

    def delete(self, key):
        wo = ldb.leveldb_writeoptions_create()
        error = ctypes.POINTER(ctypes.c_char)()
        ldb.leveldb_delete(self.db, wo, key, len(key), ctypes.byref(error))
        ldb.leveldb_writeoptions_destroy(wo)
        _checkError(error)

    def iterate(self, start=None, end=None):
        ro = ldb.leveldb_readoptions_create()
        it = ldb.leveldb_create_iterator(self.db, ro)
        ldb.leveldb_readoptions_destroy(ro)
        if start is None:
            ldb.leveldb_iter_seek_to_first(it)
        else:
            ldb.leveldb_iter_seek(it, start, len(start))
        try:
            while ldb.leveldb_iter_valid(it):
                size = ctypes.c_size_t(0)
                keyPtr = ldb.leveldb_iter_key(it, ctypes.byref(size))
                key = ctypes.string_at(keyPtr, size.value)
                if end is not None and key >= end:
                    break
                valPtr = ldb.leveldb_iter_value(it, ctypes.byref(size))
                val = ctypes.string_at(valPtr, size.value)
                yield key, val
                ldb.leveldb_iter_next(it)
        finally:
            ldb.leveldb_iter_destroy(it)

    def keys(self):
        ro = ldb.leveldb_readoptions_create()
        it = ldb.leveldb_create_iterator(self.db, ro)
        ldb.leveldb_readoptions_destroy(ro)
        ldb.leveldb_iter_seek_to_first(it)
        try:
            while ldb.leveldb_iter_valid(it):
                size = ctypes.c_size_t(0)
                keyPtr = ldb.leveldb_iter_key(it, ctypes.byref(size))
                key = ctypes.string_at(keyPtr, size.value)
                yield key
                ldb.leveldb_iter_next(it)
        finally:
            ldb.leveldb_iter_destroy(it)



import math
import numpy
from amulet.api.data_types import FloatTriplet, PointCoordinates


def scale_matrix(sx: float, sy: float, sz: float) -> numpy.ndarray:
    return numpy.array(
        [[sx, 0, 0, 0], [0, sy, 0, 0], [0, 0, sz, 0], [0, 0, 0, 1]], dtype=numpy.float64
    )


def displacement_matrix(x: float, y: float, z: float) -> numpy.ndarray:
    return numpy.array(
        [[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [x, y, z, 1]], dtype=numpy.float64
    )


def rotation_matrix(*angles, order="xy") -> numpy.ndarray:
    mat = numpy.identity(4, dtype=numpy.float64)

    for angle, axis in zip(angles, order):
        c = math.cos(angle)
        s = math.sin(angle)
        if axis == "x":
            mat = numpy.matmul(
                numpy.array(
                    [[1, 0, 0, 0], [0, c, s, 0], [0, -s, c, 0], [0, 0, 0, 1]],
                    dtype=numpy.float64,
                ),
                mat,
            )
        elif axis == "y":
            mat = numpy.matmul(
                numpy.array(
                    [[c, 0, -s, 0], [0, 1, 0, 0], [s, 0, c, 0], [0, 0, 0, 1]],
                    dtype=numpy.float64,
                ),
                mat,
            )
        elif axis == "z":
            mat = numpy.matmul(
                numpy.array(
                    [[c, s, 0, 0], [-s, c, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]],
                    dtype=numpy.float64,
                ),
                mat,
            )

    return mat


def transform_matrix(
    location: PointCoordinates, scale: FloatTriplet, rotation: FloatTriplet, order="xyz"
):
    scale_transform = scale_matrix(*scale)
    rotation_transform = rotation_matrix(*rotation, order=order)
    displacement_transform = displacement_matrix(*location)
    return numpy.matmul(
        numpy.matmul(scale_transform, rotation_transform), displacement_transform
    )



from __future__ import annotations

from amulet.world_interface.chunk.interfaces.leveldb.leveldb_18.interface import (
    LevelDB18Interface,
)


class LevelDB19Interface(LevelDB18Interface):
    def __init__(self):
        super().__init__()

        self.features["chunk_version"] = 19


INTERFACE_CLASS = LevelDB19Interface



from __future__ import annotations

from amulet.world_interface.chunk.interfaces.leveldb.leveldb_20.interface import (
    LevelDB20Interface,
)


class LevelDB21Interface(LevelDB20Interface):
    def __init__(self):
        super().__init__()

        self.features["chunk_version"] = 21


INTERFACE_CLASS = LevelDB21Interface



from __future__ import annotations

from amulet.world_interface.chunk.interfaces.anvil.anvil_1444_1_13.anvil_1444_1_13_interface import (
    Anvil1444Interface,
)


class Anvil1466Interface(Anvil1444Interface):
    def __init__(self):
        Anvil1444Interface.__init__(self)
        self.features["height_map"] = "C|36LA|V1"

    @staticmethod
    def minor_is_valid(key: int):
        return 1466 <= key < 1467


INTERFACE_CLASS = Anvil1466Interface
